/*
 * In dieser Datei sind allgmeine Variablen, Datentypen und Funktionen
 * definiert, die auch zur Implementierung von anderen Zellularautomaten
 * nützlich sind.
 */

// Hier wird der "Einstiegspunkt" für das Programm festgelegt: Sobald das
// HTML-Dokument vollständig geladen ist, löst der Browser den
// "DOMContentLoaded"-Event aus, der hier mit dem Aufruf der Funktion
// initialize() verknüpft wird:
document.addEventListener("DOMContentLoaded", initialize, false);

// Die Variable Settings gruppiert alle globalen Einstellungen. Der Großteil der
// Werte wird zunächst auf '0' (bzw. je nach Funktion auf entsprechende
// Standardwerte) initialisiert. Die eigentlichen Einstellungen werden erst
// durch die Funktion initialize() gesetzt:
var Settings = {
    boardDivision: 0,        // Enthält später die Zellunterteilung in Pixeln.
    boardSideLenght: 0,      // Enthält später die Seitenlänge des Spielfelds in
                             // Pixeln.
    canvas: undefined,       // Enthält später die Referenz auf den Canvas im
                             // HTML-Dokument.
    context: undefined,      // Enthält später die Referenz auf den
                             // Zeichenkontext im Canvas, damit JavaScript weiß,
                             // wohin es zeichnen soll.
    cellsPerLine: 50,        // Teilt den canvas in 50 Zellen pro Zeile.
    cellSize: 0,             // Enthält später die tatsächliche Größe einer
                             // Zelle (in Pixeln), benötigt zum Zeichnen.
    cellColor: ["#ffffff"],  // Diese Liste wird in wireworld.js mit den übrigen
                             // Zellfarben ergänzt.
    cellToDraw: 1,           // Hier wird der aktuell zu zeichnende Zelltyp
                             // festgelegt.
    intervalID: 0,           // Enthält später eine Referenz auf das Intervall-
                             // Timerobjekt, das den Ablauf des Automaten
                             // steuert.
    speed: 100               // Speichert die aktuelle Geschwindigkeit der
                             // Simulation (in ms).
};


// Hier wird Cell als ein eigener Datentyp definiert, der eine einzelne Zelle
// repräsentiert. So kann später für jede Zelle sowohl ihr Status als auch die
// Anzahl ihrer lebenden Nachbarn gespeichert werden. Beide Attribute werden mit
// '0' initialisiert. Der Status '0' enspricht in der Game-of-Life-Terminologie
// einer 'toten' Zelle, '0' Nachbarn sind selbsterklärend.
var Cell = function ()
{
    this.state = 0;
    this.neighbours = 0;
}


// Field ist der Datentyp für das Spielfeld. Mit der folgenden Konstruktor-
// Funktion wird er initialisiert:
var Field = function (rows, columns)
{
    this.rows  = rows;                // Die Attribute rows und colums werden
    this.columns = columns;           // für diverse Schleifen als Maximalwerte
                                      // benötigt und deswegen hier zentral
                                      // gespeichert.
    this.generation = 0;              // Enthält die aktuelle Generation.
    this.generationHTML = undefined;  // Wird durch initialize() gesetzt,
                                      // verweist dann auf die
                                      // Generationenanzeige im HTML-Dokument.
    this.population = 0;              // Enthält die aktuelle Population (in
                                      // Wireworld: Die Anzahl der
                                      // Elektronenköpfe.)
    this.oldPopulaton = 0;
    this.populationHTML = undefined;  // Wird durch initialize() gesetzt,
                                      // verweist dann auf die
                                      // Populationsanzeige im HTML-Dokument.
    this.data = [];                   // Der eigentliche Spielfeld-Array.

    // Fülle den Spielfeld-Array mit Cell-Objekten. Zur Vereinfachung der
    // Regelanwendung wird eine zusätzliche Zeile außen um das sichtbare
    // Spielfeld erzeugt, daher 'i < this.rows + 2', bzw. 'i < this.coloumns +
    // 2':
    for (var i = 0; i < this.rows + 2; i++) {
        // Jede Zeile ist selbst ein Array, sodass wir insgesamt ein
        // zweidimensionales Array erhalten:
        this.data[i] = [];

        for (var j = 0; j < this.columns + 2; j++)
            // Hier werden die eigentlichen Cell-Objekte mittels der
            // Arary-Methode push() erzeugt und angehängt:
            this.data[i].push(new Cell());
    }
}

// updateNeighbours() verändert den Nachbarzähler einer Zelle, indem der
// zweidimensionale Zellenarray rund um die eigenen Koordinaten entsprechend der
// Moore-Nachbarschaft durchlaufen wird. Sie muss jedesmal aufgerufen werden,
// wenn sich der Status einer Zelle entsprechend ändert. Durch den dritten
// Parameter 'increment' ist die Funktion allgemein einsetzbar, sowohl nach dem
// Zeichnen einer neuen, wie auch dem Entfernen einer bereits vorhandenen
// Zelle. Hier muss für increment lediglich '-1' übergeben werden.
Field.prototype.updateNeighbours = function (row, column, increment)
{
    this.data[row - 1][column - 1].neighbours += increment;
    this.data[row - 1][column].neighbours     += increment;
    this.data[row - 1][column + 1].neighbours += increment;
    this.data[row][column - 1].neighbours     += increment;
    this.data[row][column + 1].neighbours     += increment;
    this.data[row + 1][column - 1].neighbours += increment;
    this.data[row + 1][column].neighbours     += increment;
    this.data[row + 1][column + 1].neighbours += increment;
}

// resetAllNeighbours() wird von der nextGeneration()-Funktion aus aufgerufen,
// um anschließend, d.h. nach der Regelanwendung, alle Nachbarn neu zu zählen:
Field.prototype.resetAllNeighbours = function()
{
    for (var row in this.data)
        for (var column in this.data[row])
            this.data[row][column].neighbours = 0;
}

// increaseGeneration() erhöht den Generationszähler und aktualisiert die
// Anzeige im HTML-Dokument:
Field.prototype.increaseGeneration = function()
{
    this.generation++;
    this.generationHTML.innerHTML = this.generation;
}

// updatePopulationHTML() aktualisiert die Anzeige der Population. In Wireworld
// gibt die Population die Anzahl der Elektronenköpfe wieder:
Field.prototype.updatePopulationHTML = function ()
{
    // Aktualsiere nur, wenn sich an der absoluten Anzahl etwas geändert hat!
    if (this.population != this.oldPopulation) {
        this.populationHTML.innerHTML = this.population;
        this.oldPopulation = this.population;
    }
}

// clear() setzt alle Zellen auf Status '0', also 'leer' zurück. Wird nach
// Nutzerklick auf den 'Clear Field'-Button aufgerufen:
Field.prototype.clear = function ()
{
    for (var row = 1; row <= this.rows; row++) {
        for (var column = 1; column <= this.columns; column++) {
            this.createCell(row, column, 0, -1);
            drawCell(row, column, 0);
            this.updatePopulationHTML();
        }
    }

    // Generationszähler erst intern zurücksetzen und anschließend im HTML-
    // Dokument übernehmen:
    this.generation = 0;
    this.generationHTML.innerHTML = this.generation;
}

// initialize() ist der Einstiegspunkt des Programms: HTML-Elemente ermitteln,
// die Canvas-Zeichenoberfäche vorbereiten und EventListener für Nutzerintaktion
// setzen (Mausklicks zum Zeichnen von Zellen sowie die 'DropZone' für das Laden
// von gespeicherten Zellkonfigurationen):
function initialize()
{
    // HTML-Elemente ermitteln:
    Settings.canvas = document.getElementById("board");
    Settings.context = Settings.canvas.getContext("2d");
    field.generationHTML = document.getElementById("Generation");
    field.populationHTML = document.getElementById("Population");

    // Zeichenoberfläche vorbereiten. Es wird eine quadratische Grundstruktur
    // angenommen:
    Settings.boardSideLength = Settings.canvas.width;
    // Durch abrundendes Teilen der Seitenlänge des Canvas' durch die Anzahl der
    // zu zeichnenden Zellen wird die Unterteileilung des Canvas' (in Pixel)
    // ermittelt:
    Settings.boardDivision = Math.floor(Settings.boardSideLength /
                                        Settings.cellsPerLine);
    // Von dieser boardDivision wird die Breite des noch zu zeichnenden
    // Gitternetzes subtrahiert, um die tatsächliche Zellenbreite zu erhalten:
    Settings.cellSize = Settings.boardDivision - 2;

    // Nun kann das Gitternetz gezeichnet werden:
    Settings.context.strokeStyle = "#cccccc";  // Farbe des Gitternetzes
    Settings.context.lineWidth = 2;            // Linienbreite in Pixeln

    // Pro Schleifendurchlauf wird je eine horizontale und eine vertikale Linie
    // gezeichnet:
    for (var i = 0; i <= Settings.boardSideLength;
         i = i + Settings.boardDivision) {
        Settings.context.moveTo(0, i);
        Settings.context.lineTo(Settings.boardSideLength, i);
        Settings.context.moveTo(i, 0);
        Settings.context.lineTo(i, Settings.boardSideLength);
        Settings.context.stroke();
    }

    // EventListener für Mausklicks auf dem Canvas:
    Settings.canvas.addEventListener("mousedown", drawCellFromMouseClick,
                                     false);

    // EventListener, um den Canvas zu einer 'DropZone' für das Laden von
    // Zellkonfigurationen zu machen:
    var dropZone = Settings.canvas;
    dropZone.addEventListener("dragenter", dragenter, false);
    dropZone.addEventListener("dragover", dragover, false);
    dropZone.addEventListener("drop", drop, false);
}



/*
 * Ende des Codes für die grundsätzliche Zellularautomatenfunktionalität.
 *
 * Ab hier folgt nur noch Code, der ausschließlich für die Nutzerinteraktion
 * benötigt wird.
 */

// drawCellFromMouseClick() ist die Callback-Funktion, wenn der Nutzer innerhalb
// des Canvas' einen Mausklick ausführt. Die Koordinaten des Mausklicks werden
// ermittelt, auf die entsprechenden Spielfeldkoordinaten umgerechnet und dort
// eine Zelle gezeichnet.
//
// Der Code wurde adaptiert von:
// http://miloq.blogspot.de/2011/05/coordinates-mouse-click-canvas.html
function drawCellFromMouseClick(event)
{
    // Ermittle die absoluten Koordinaten des Mausklicks (bezogen auf das
    // gesamte Browserfenster):
    var x = event.x;
    var y = event.y;

    // Firefox unterstützt event.x und event.y nicht, folgender Workaround
    // funktioniert jedoch analog:
    if (x == undefined && y == undefined) {
        x = event.clientX;
        y = event.clientY;
    }

    // Wenn die Ansicht im Browserfenster durch Scrollen bewegt wurde, muss
    // dieser Versatz mit einberechnet werden:
    x += window.pageXOffset;
    y += window.pageYOffset;

    // Abzüglich des Pixel-Koordinaten der oberen linken Ecke des Canvas'
    // erhalten wir die korrekten Klick-Koordinaten innerhalb des Canvas':
    x -= Settings.canvas.offsetLeft;
    y -= Settings.canvas.offsetTop;

    // Übersetzung der Pixel-Koordinaten in die entsprechenden Spielfeld-
    // Koordinaten (row, column):
    var coords = convertClickToFieldCoordinates(x, y);
    var row    = coords[0];
    var column = coords[1];

    // Versuche, an den ermittelten Koordinaten eine Zelle zu erstellen. Wenn
    // dies gelingt, zeichne die entsprechende Zelle auf den Canvas. Wenn es
    // jedoch misslingt, weil sich die dort befindliche Zelle bereits im
    // gewünschten Zustand befindet, gibt field.createCell() false zurück und
    // der else-Fall tritt ein. Der Mausklick 'löscht' dann die an den
    // Koordinaten befindliche Zelle, indem die dortige Zelle auf dem Status
    // '0', bzw. 'leer' gesetzt wird.
    if (field.createCell(row, column, Settings.cellToDraw, 1))
        drawCell(row, column, Settings.cellToDraw);
    else {
        field.createCell(row, column, 0, -1);
        drawCell(row, column, 0);
    }
}

// convertClickToFieldCoordinates() rechnet pixelbasierte Klick-Koordinaten in
// die entsprechenden Spielfeld-Koordinaten um:
function convertClickToFieldCoordinates(x, y) {
    // Math.floor() rundet auf ganze Zahlen ab, das '+ 1' am Ende berücksichtigt
    // die 'unsichtbaren' Randzellen, die rund um das sichtbare Spielfeld
    // liegen.
    var row = Math.floor(y / Settings.boardDivision) + 1;
    var column = Math.floor(x / Settings.boardDivision) + 1;

    // Wird ganz am rechten, bzw. am unteren Rand des Canvas' geklickt, können
    // Koordinaten erzeugt werden, die außerhalb des gültigen Bereiches
    // liegen. Wenn dies der Fall ist, reduziere die Koordinaten entsprechend.
    if (row > field.rows)
        row--;

    if (column > field.columns)
        column--;

    return [row, column];
}

// drawCell() zeichnet eine Zelle auf den Canvas:
function drawCell(row, column, state)
{
    // context.fillStyle bestimmt die Zeichenfarbe. Setze sie entsprechend des
    // zu malenden Zellstatus (die entsprechenden Farbdefinitionen finden sich
    // zu Beginn der wireworld.js):
    Settings.context.fillStyle = Settings.cellColor[state];

    // Die Parameter der Funktion sind Feldkoordinaten, wir benötigen jetzt die
    // entsprechenden Pixelkoordinaten auf dem Canvas. Im Prinzip die inverse
    // Funktion zu convertClickToFieldCoordinates() oben. Das '+ 1' sorgt hier
    // dafür, dass das Zellgitter nicht übermalt wird:
    var drawx = Math.floor(column * Settings.boardDivision)
        - Settings.boardDivision + 1;
    var drawy = Math.floor(row * Settings.boardDivision)
        - Settings.boardDivision + 1;

    // Zeichne das eigentliche Quadrat:
    Settings.context.fillRect(drawx, drawy, Settings.cellSize,
                              Settings.cellSize);
}

// clearField() wird aufgerufen, wenn der Nutzer auf den 'Clear Field'-Button
// klickt.
function clearField()
{
    // Es wird lediglich auf die Methode field.clear() umgeleitet:
    field.clear();
}

// runAutomata() startet die automatische Simulation des zellulären Automaten.
function runAutomata()
{
    // Wenn der Automat noch nicht läuft starte einen regelmäßigen Timer, der
    // alle <Settings.speed> Millisekunden die Funktion
    // nextGenerationFromInterval() aufruft.
    if (!Settings.intervalID)
        // Die Referenz auf diesen Timer (die intervalID) wird in den Settings
        // gespeichert, um ihn später wieder stoppen zu können:
        Settings.intervalID = window.setInterval(nextGenerationFromInterval,
                                                 Settings.speed);
}

// stopAutomata() stoppt die Simulation wieder.
function stopAutomata()
{
    // Stoppe den Timer und lösche die Referenz aus den Settings:
    window.clearInterval(Settings.intervalID);
    Settings.intervalID = 0;
}

// changeSpeed() wird aufgerufen, wenn der Nutzer ein anderes Element aus der
// Geschwindigkeits-Auswahlliste auswählt.
function changeSpeed(selectItem)
{
    // Speichere den neu gewählten Wert:
    Settings.speed = selectItem.value;

    // Falls die automatische Simulation bereits läuft, stoppe kurz und starte
    // erneut mit der neuen Geschwindigkeit:
    if (Settings.intervalID) {
        stopAutomata();
        runAutomata();
    }
}

// nextGeneration() wird aufgerufen, wenn der Nutzer auf den 'Next Generation'-
// Button klickt.
function nextGeneration()
{
    // Berechne nur dann die nächste Generation, wenn die automatische
    // Simulation gestoppt ist (Settings.intervalID ist dann = 0), andernfalls
    // stoppe den Automaten, weil der Nutzer nun offenbar selbst über die
    // Generationenschritte entscheiden möchte:
    if (!Settings.intervalID)
        field.nextGeneration();
    else
        stopAutomata();
}

// Callback-Funktion für den Timer.
function nextGenerationFromInterval()
{
    // Es wird auf field.nextGeneration umgeleitet:
    field.nextGeneration();
}


/*
 * saveFile() und loadFile() ermöglichen das Speichern und Laden von
 * Zellkonfigurationen in und aus einfachen Textdateien.
 *
 * Der Code für die beiden Funktionen wurde adaptiert von:
 * https://thiscouldbebetter.wordpress.com/2012/12/18/loading-editing-and-
 * saving-a-text-file-in-html5-using-javascrip/
 */

// saveFile() speichert die aktuelle Zellkonfiguration auf dem Canvas in einer
// Textdatei. Da JavaScript im Webbrowser nur eingeschränkte Rechte besitzt,
// z.B. nicht schreibend auf das Dateisystem zugreifen darf, was nötig wäre, um
// wie in einer normalen Desktop-Applikation einen 'Datei speichern'-Dialog
// anzeigen und die Datei dann auch tatsächlich abspeichern zu können, muss hier
// getrickst werden. Es wird ein zusätzlicher (unsichtbarer) Downloadlink
// erzeugt, der von JavaScript selbst "angeklickt" wird und anschließend wieder
// entfernt wird. Ähnlich funktioniert auch das von diversen Downloadseiten im
// Internet bekannte "Ihr Download startet automatisch in 5 Sekunden..."
function saveFile()
{
    // Erzeuge die zu speichernden Textdaten aus dem aktuellen Spielfeld:
    var textToWrite = '';

    // Beginnend bei (1,1), also nur den im Canvas sichtbaren Teil:
    for (var row = 1; row <= field.rows; row++) {
        for (var column = 1; column <= field.columns; column++) {
            // field.generateChar() gibt das zum Zellstatus gehörende ASCII-
            // Zeichen zurück. Die Definition befindet sich in wireworld.js.
            textToWrite += field.generateChar(row, column);
        }
        // Beende jede Zeile mit einem 'New line'-Zeichen. Dies wäre zwar nicht
        // unbedingt nötig, sorgt aber dafür, dass die erzeugte Textdatei
        // menschenlesbar ist und z.B. von Hand in einem Texteditor bearbeitet
        // werden kann:
        textToWrite += '\n';
    }

    // Der Inhalt des Textfeldes 'FileName' dient als Dateiname:
    var fileName = document.getElementById("FileName").value;

    // Ab hier beginnt die Trickserei: Aus dem oben erzeugten textToWrite wird
    // ein temporäres Blob-Objekt erzeugt, das wie eine normale Textdatei auf
    // der Festplatte aussieht. (Blob ist Teil der File API, die aktuell von der
    // W3C für die Interaktion mit Dateien in JavaScript entwickelt wird, siehe:
    // http://www.w3.org/TR/FileAPI/
    var textFileAsBlob = new Blob([textToWrite], {type:'text/plain'});

    // Nun wird der "unsichtbare" Download-Link erzeugt:
    var downloadLink = document.createElement("a");
    downloadLink.download = fileName;

    // Die folgende if-Abfrage ist nötig, da die aufzurufende Funktion in
    // verschiedenen Browsern unterschiedliche Namen hat:
    if (window.webkitURL != null) {
        // Webkit-basierte Broser, wie Google Chrome oder Apple Safari:
        downloadLink.href = window.webkitURL.createObjectURL(textFileAsBlob);
    } else {
        // Mozilla Firefox:
        downloadLink.href = window.URL.createObjectURL(textFileAsBlob);

        // Firefox benötigt einen zusätzlichen Workaround gegenüber Webkit-
        // Browsern. Hier muss der Link tatsächlich zum DOM (Document Object
        // Model, der browserinternen Repräsentation des HTML-Dokuments)
        // hinzugefügt werden, bevor er von JavaScript "angeklickt" werden kann.
        // Die Funktion onclick() sorgt dafür, dass dies nach dem simulierten
        // Klick sofort wieder rückgängig gemacht wird:
        downloadLink.onclick = function (e) {
            document.body.removeChild(e.target);
        };
        downloadLink.style.display = "none";     // Mach den Link unsichtbar und
        document.body.appendChild(downloadLink); // hänge ihn ans Dokument an.
    }

    // Simuliere einen Klick auf den erzeugten Link, um den Download auszulösen.
    downloadLink.click();
}

// loadFile() lädt eine entsprechend formatierte Textdatei und parst ihren
// Inhalt in den Spielfeld-Canvas.
function loadFile(files)
{
    // Wähle nur die erste Datei aus der files-Liste aus (files können mehrere
    // Dateien seien, wenn der Nutzer mehrere markiert und gleichzeitig in den
    // Canvas gezogen hat).:
    var file = files[0];

    // Regulärer Ausdruck, der auf die MIME-Typen von Textdateien passt:
    var textType = /text.*/;

    // Fahre nur fort, wenn es sich wirklich um eine Textdatei handelt:
    if (file.type.match(textType)) {
        // Erzeuge ein FileReader-Objekt (Teil der File API, siehe den Link oben
        // zu Blob):
        var reader = new FileReader();

        // Die onload()-Funktion wird aufgerufen, wenn der Browser die Datei
        // fertig von der Festplatte gelesen hat.
        reader.onload = function (e) {
            field.clear();

            // Wir laden nur in den sichtbaren Teil des Spielfelds, deswegen
            // beginnen wir bei (1,1):
            var row = 1, column = 1;

            // reader.result enthält den Inhalt der Textdatei in einem Array,
            // Zeichen für Zeichen. Mit einer for-Schleife können wir also --
            // Zeichen für Zeichen -- mit dem Parsen fortfahren:
            for (var charNr in reader.result) {
                // field.parseChar() erzeugt eine dem ASCII-Zeichen
                // entsprechende Zelle auf dem Canvas. Die Definition findet
                // sich in wireworld.js:
                field.parseChar(row, column, reader.result[charNr]);

                // Erhöhe den Spaltenzähler:
                column++;

                // Wenn dies die letzte Spalte war, setze den Spaltenzähler auf
                // '0' zurück und erhöhe den Zeilenzähler:
                if (column > field.columns) {
                    column = 0;
                    row++;
                }
            }

            // Zähle alle Nachbarn und aktualisiere den Populationszähler.
            field.countAllNeighbours();
            field.updatePopulationHTML();
        };

        // Nachdem das FileReader-Objekt um die nötige Lesefunktion onload
        // ergänzt wurde, starte nun den eigentlichen Lesevorgang von der
        // Festplatte.
        reader.readAsText(file);
    } else {
        console.log("File not supported.")
    }
}

// Callback-Funktionen für die 'DropZone' auf dem Canvas. dragenter() und
// dragover() tun nichts. In diesen Funktionen wird nur verhindert, dass
// evtl. im Browser definierte Standardaktionen für die genannten Ereignisse
// ausgeführt werden.
function dragenter(e)
{
    e.stopPropagation();
    e.preventDefault();
}

function dragover(e)
{
    e.stopPropagation();
    e.preventDefault();
}

// drop() gibt die gewünschte(n) Datei(en) an die loadFile()-Funktion weiter:
function drop(e)
{
    e.stopPropagation();
    e.preventDefault();

    loadFile(e.dataTransfer.files);
}

// handleKeyPress() vereinfacht die Nutzerinteraktion: Befindet sich der Nutzer
// mit seinem Text-Cursor innerhalb eines bestimmten Textfeldes, kann mit der
// Returntaste eine Funktion ausgelöst werden.
function handleKeyPress(event, textField)
{
    if (event.keyCode === 13) {  // '13' ist der keyCode für die 'Return'-Taste.
        // Wenn im textField namens 'FileName', rufe die Speicherfunktion auf:
        if (textField == "FileName")
            saveFile();
        // Wenn im textField 'RandomCells', fülle den Canvas zufällig. (In
        // Wireworld ohne Bedeutung, Überbleibsel des Life-Automaten)
        else if (textField == "RandomCells")
            fillWithRandomCells();
    }
}
